/*
* ----------------------------------------------------------------------------
* "THE BEER-WARE LICENSE" (Revision 42):
* Rizo (Richard) Deininger (richard.deininger@lieberlieber.com) wrote this file. As long as you retain this notice you can do whatever you want with this stuff. 
* If we meet some day, and you think this stuff is worth it, you can buy me a beer in return.
*
* Disclaimer of Warranty.
* THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY APPLICABLE LAW. EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER 
* PARTIES PROVIDE THE PROGRAM “AS IS” WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF  
* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM IS WITH YOU. SHOULD THE PROGRAM PROVE  
* DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR CORRECTION.
*
* Limitation of Liability. 
* IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS THE PROGRAM  
* AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OR INABILITY TO  
* USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A FAILURE OF THE  
* PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS), EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
* ----------------------------------------------------------------------------
*/


#include <DHT.h>
#include <Regexp.h>
#include <EEPROM.h>
#include <SoftwareSerial.h>
SoftwareSerial SIM900(3, 2);

// EEPROM address storage
const int NumberStartAddr = 0;
const int NumberEndAddr = 19;
const int TempAddr = 20;

// AT commands and answers
const String SMS_Header = "+CMT: \"";

const String COM_SmsTextMode = "AT+CMGF=1";
const String COM_SmsNotification = "AT+CNMI=2,2,0,0";
const String COM_DeleteReadSMS = "AT+CMGD=1,4";
const String COM_SendSMS = "AT+CMGS=\"";
const String COM_SleepMode = "AT+CSCLK=2"; // 0 = disabled, 1 = active, 2 = auto
const String COM_EchoOff = "ATE0";

// AT command confirmation tag
const String CONFIRMATION = "OK|";


// SMS keywords
const String KEY_NewTemp = "new";
const String KEY_AlterNumber = "alarm";
const String KEY_CurrentTemp = "temp";

//User interaction strings 
const String TXT_Ready = "Ready";
const String TXT_NewAlertNum = "New alert number: ";
const String TXT_NewAlertTemp = "New alert temp.: ";
const String TXT_SensorError = "Temperature sensor error!";
const String TXT_Temp = "Temperature: ";
const String TXT_Humid = "Humidity: ";
const String TXT_Alert = "ALERT!";

  
// default settings
String alertNumber = "+43123456789"; // !!! TODO: set real default number bevore deploy !!!
float alertTemp = 30;

// saves last sent temperature (see CheckTemp function)
float lastAlertTemp = 0;

// Temperature pin settings
#define DHTPIN 13
#define DHTTYPE DHT22 
DHT dht(DHTPIN, DHTTYPE);



void setup() 
{
   
   Serial.begin(4800); // for serial monitor
   SIM900.begin(4800); // for GSM shield
   
  // read alert number from EEPROM
  String tmpNum = GetNumber();
  if (tmpNum != "")
  {
    alertNumber = tmpNum;      
  }

  // read alert temperature from EEPROM  
  if (GetTemp() != 0)
  {
    alertTemp = GetTemp();
  }
  
  
  delay(10000);  // give time to log on to network.  
  
  // check if SIM is ready
  SendCommandWaitConfirm("AT");
  // Set SIM to inital configuration
  SendCommandWaitConfirm(COM_EchoOff);
  SendCommandWaitConfirm(COM_SmsNotification);
  SendCommandWaitConfirm(COM_SmsTextMode);

  // Show/Send wakeup confirmation
  Serial.println("Up and running");     
  SendSMS(alertNumber, TXT_Ready+"\r\n"+GetTempData());
}

/*
  Main loop. 
  Checks incomming messages and if alarm temp was reached
*/
void loop() {
  // put your main code here, to run repeatedly:

  // Check for incomming messages/AT commands
  Read();
  delay(500);
  // check if additional data is available from SIM if not sleep for 5 seconds
  if (!SIM900.available() > 0)
  {
    delay(5000);
    // Check if alarm temp. is reached
    CheckTemp();
  }  
}

/*
  Reads incomming AT commands and checks if it contains SMS header
*/
void Read()
{
    String message = ReadSim(); //Read Data form Sim
    // discard empty messages
    if (message != "")
    {
      Serial.println(message);
      ReadHeader(message); // Checks Header for SMS tag
    }
}

/*
  Checks message from SIM for SMS header
  SMS message example: 
      +CMT: "+0123456789",,"09/10/26,16:00:44+00" |SMS text is here|
    
*/
void ReadHeader(String message)
{ 
   
   if (message.startsWith(SMS_Header))
   {
      String number = "";
      String smsText = "";
      int lastIndex = 0;  
      String tmp = message;

      // check if message contains more then one SMS Header tag
      while (message.indexOf(SMS_Header, lastIndex+ SMS_Header.length()) > -1)
      {
          delay(1000);
          // read first SMS (from header one to header two)
          tmp = message.substring(lastIndex,message.indexOf(SMS_Header, lastIndex+ SMS_Header.length()));    
          
          // get phone number and text from SMS
          number = ExtractNumber(tmp);
          smsText = ExtractSMStext(tmp);
          
          // if Number and Text was found check if message contains keywords
          if (number != "" && smsText != "")
          {
            CheckSMSText(number, smsText);
          }

          // select next header to read
          lastIndex = message.indexOf(SMS_Header, lastIndex+ SMS_Header.length());
          
          // if no more text is found stop searching
          if (tmp == "")
          {
            break;
          }
      }

      // read only or last SMS
      tmp = message.substring(lastIndex,message.length());
      
     // get phone number and text from SMS
      number = ExtractNumber(tmp);
      smsText = ExtractSMStext(tmp);
       // if Number and Text was found check if message contains keywords
      if (number != "" && smsText != "")
      {
        CheckSMSText(number, smsText);
      }
      
      // delete all SMS from SIM
      DeleteAllSMS();
   }
}

/*
  Sends a SMS to the provided number with the text provided
  
    
*/
void SendSMS(String number, String text)
{
  Serial.println("Sending: "+text+" to: "+number);
  String Answer = "";
  
  // resend SMS until SIM sends confirmation message back
  while (!Answer.endsWith(CONFIRMATION))
  {
    // Sends SMS
    delay(200);
    SIM900.print("AT+CMGS=\""+number+"\"\r");
    delay(1500);        
    SIM900.print(text+"\r");
    delay(1500);
    SIM900.write(0x1A);  //Equivalent to sending Ctrl+Z     
    
    // Reads answer from SIM
    delay(1000);  
    Answer = ReadSim();
    Serial.println("Answer: "+Answer);
    // debug message if no confirmation was sent from SIM
    if (!Answer.endsWith(CONFIRMATION))
    {
      Serial.println("Resending SMS");
    }
  }
}

/*
  Extracts the number of the sender from a SMS
  SMS message example: 
      +CMT: "+0123456789",,"09/10/26,16:00:44+00" |SMS text is here|
    
*/
String ExtractNumber(String message)
{  
  // convert to CharArray to be used by Regex
  char charBuf[message.length()];
  message.toCharArray(charBuf, message.length());
  MatchState ms; 
  ms.Target(charBuf);
  String number = ""; // set empty number for validation check

  // extracting number
  if (ms.Match ("[+]CMT: \"[0-9+]*\"") > 0) 
  {    
    number = message.substring(SMS_Header.length(), ms.MatchLength-1);      
  }
  return number;
}

/*
  Extracts the text form a SMS
  SMS message example: 
      +CMT: "+0123456789",,"09/10/26,16:00:44+00" |SMS text is here|
  
*/
String ExtractSMStext(String message)
{
  
  String text = ""; // set empty text for validation check
  // searching for delimiter | to extract SMS text
  if (message.endsWith("|"))
  {
    String tmp = message.substring(0, message.length()-1); // removing last delimiter
    if (tmp.lastIndexOf("|") > -1) // searching next delimiter (from end since the SMS text is the last text
    {
      text = tmp.substring(tmp.lastIndexOf("|")+1);
    }
  }
  return text;
}


/*
  Reads data sent from SIM to Arduino
  all NL and CR are replaced with a custom delimiter |
*/
String ReadSim()
{
   String inputString= "";
   bool stringComplete = false;
   bool lastWasDelimiter = false;
  
  //  Read until the message end
   do
   {  
     delay(1000);
     // read all available data
     while (SIM900.available() > 0)    
     {      
        // get the new byte:
        char inChar = (char)SIM900.read(); 

        // add it to the inputString, filter all "unknown" characters (sometimes when the battery is low SIM sends: ÿ == chr(255))
        if (inChar > 31 && inChar < 128)
        {
          inputString += inChar;
          lastWasDelimiter = false;
        } else if ((inChar == 13 || inChar == 10) && !lastWasDelimiter && inputString != "") // replace NL and CR, discard double delimiters and empty strings
        {
            inputString += '|';    
            lastWasDelimiter = true;
        }      
     }    
   } while (!inputString.endsWith("|") && inputString != "");
   
   return inputString;   
}

/*
  Sends a command to the SIM and waits for confirmation
  if not confirmed from SIM: wait 2 sec. and resend
*/
void SendCommandWaitConfirm(String command)
{
  String Answer = "";
  // do this until SIM answers with confirmation
  while (!Answer.endsWith(CONFIRMATION))
  {
    // sending AT command to SIM
    Serial.println("Sending Command: "+command);
    SendCommand(command);
    // Wait and read answer
    delay(2000);
    Answer = ReadSim();
    
    //debug message of answer and possible resend
    Serial.println("Answer: "+Answer);
    if (!Answer.endsWith(CONFIRMATION))
    {
      Serial.println("Resending Command");
    }
  }
    
}

/*
  Sends AT delete command to SIM
*/
void DeleteAllSMS()
{
  SendCommandWaitConfirm(COM_DeleteReadSMS);
}

/*
  Sends AT command to SIM and appends it with a CR for termination
*/
void SendCommand(String command)
{
  SIM900.print(command+"\r");
}


/*
  checks SMS text for keywords
*/
void CheckSMSText(String number, String message)
{
    message.toLowerCase(); // first convert all to lower case
    
    if (message.indexOf(KEY_AlterNumber) > -1) // check if alarm number needs to be changed
    {
       // Save in EEPROM for future set new alert number and send confirmation to user
       SaveNumber(number);
       alertNumber = number;
       SendSMS(number, TXT_NewAlertNum+ number);      
    } else if (message.indexOf(KEY_NewTemp)  > -1) // check if new temperature needs to be set
    {
      message = message.substring(KEY_NewTemp.length()); // remove keyword
      message = message.substring(0,message.indexOf("o")); // remove degree char (not needed)
      message.trim(); // remove any spaces

      // convert, set and save new temperature to EEPROM
      int temp = message.toInt();
      alertTemp = temp;     
      SaveTemp(temp);
      // Send confirmation to user
      SendSMS(number, TXT_NewAlertTemp+String(temp));
    } else if (message.indexOf(KEY_CurrentTemp)  > -1) // check if current temperature needs to be sent
    {
      // send current temperature to user
      SendSMS(number, GetTempData());
    } 

}

/*
  Reads temperature and humidity date from the DHT22 sensor
  returns data as string (for SMS sending)
*/
String GetTempData()
{
  // Read humidity
  float h = dht.readHumidity();
  // Read temperature as Celsius
  float t = dht.readTemperature();
  
  // Check if any reads failed and send sensor error to user
  if (isnan(h) || isnan(t)) 
  { 
    SendSMS(alertNumber, TXT_SensorError);
  }    
   return TXT_Humid+String(h)+"%\r\n"+TXT_Temp+String(t);
}

/*
 Reads temperature and humidity date from the DHT22 sensor
 Checks it against the alert temperature
*/
void CheckTemp()
{
  // Read humidity
  float h = dht.readHumidity();
  // Read temperature as Celsius
  float t = dht.readTemperature();
  
  // Check if any reads failed and send sensor error to user
  if (isnan(h) || isnan(t)) 
  { 
    SendSMS(alertNumber, TXT_SensorError);
  }
    
   // check if alert temperature is reached    
   if (t >= alertTemp)
   {
     // check for last alert temperature
     // temperature is checked every 5 sec. therfore we save the temperature from last alert
     // and resend an alert if the temperature change more the 4 degree
     Serial.println("ABS: "+String(abs(lastAlertTemp - t)));
     if (abs(lastAlertTemp - t) > 4)
     {
       // saving last alert temperature and sending alert message
       lastAlertTemp = t;
       SendSMS(alertNumber, TXT_Alert+"\r\n"+TXT_Humid+String(h)+"%\r\n"+TXT_Temp+String(t));                      
     }
   } else 
   {
     lastAlertTemp = 0; // resetting last alert temperature if falling below alert temperature
   }
}

/*
  Saves a new alert number to EEPROM
*/
void SaveNumber(String number)
{
  // converting number to byte array max size 20 (should be enought)
  byte buff[20];
  number.getBytes(buff, number.length()+1);

  // writing byte array to EEPROM number address store
  for (int i = NumberStartAddr; i <= NumberEndAddr; i++)
  {
    byte tmp = 0;
    if (i < number.length())
    {
      tmp = buff[i];
    }     
    EEPROM.write(i,tmp);
  }
}

/*
  Reading alert number from EEPROM
*/
String GetNumber()
{
  String number = ""; // setting empty number for later validation
  // reading single bytes from EEPROM
  for (int i = NumberStartAddr; i <= NumberEndAddr; i++)
  {
    char tmp = char(EEPROM.read(i));
    if (tmp != 0)
    {
      number += tmp;
    }
  }
  return number;
}

/*
  Saves new alarm temperature to EEPROM
*/
void SaveTemp(float temp)
{
  EEPROM.write(TempAddr, (int)temp); // we are reading the temperature as float but store only integer
}

/*
  Reads alarm temperature from EEPROM
*/
float GetTemp()
{
  return (float)EEPROM.read(TempAddr); // we are reading the temperature as float but store only integer
}

